/*
 * Copyright (C) 2011-2014 Intel Corporation. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <Yami.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void* DecodeInputHandler;

DecodeInputHandler createDecodeInput(char *fileName);

const char * getMimeType(DecodeInputHandler input);

bool decodeInputIsEOS(DecodeInputHandler input);

bool getNextDecodeUnit(DecodeInputHandler input, VideoDecodeBuffer *inputbuffer);

void releaseDecodeInput(DecodeInputHandler input);

#ifdef __cplusplus
}
#endif
